import dataclasses
import networkx as nx
from typing import Any, Dict, Iterable, List, Optional, Set

from magma.bits import Bits
from magma.bitutils import clog2
from magma.common import MroVisitor, Stack
from magma.smart.smart_bits import (
    SmartExprMeta,
    SmartExpr,
    SmartBitsExpr,
    SmartBits,
    issigned,
    SmartSignedOp,
    SmartExtendOp,
)


@dataclasses.dataclass(frozen=True)
class _Node:
    expr: SmartExpr
    children: List['_Node']

    def update(self, children: List['_Node']):
        self.children[:] = children

    __hash__ = object.__hash__


def _make_tree(expr: SmartExpr) -> _Node:
    children = list(_make_tree(arg) for arg in expr.args)
    return _Node(expr, children)


make_tree = _make_tree


class _Visitor(MroVisitor):
    def get_class(self, node: _Node) -> SmartExprMeta:
        return node.expr.__class__

    def generic_visit(self, node: _Node, *args, **kwargs):
        return list(
            self.visit(child, *args, **kwargs) for child in node.children
        )


class _Transformer(_Visitor):
    def generic_visit(self, node: _Node, *args, **kwargs):
        children = list(
            self.visit(child, *args, **kwargs) for child in node.children
        )
        node.update(children)
        return node


def _make_label(node: _Node) -> str:
    if isinstance(node.expr, SmartBitsExpr):
        return str(node.expr)
    label = str(node.expr)
    try:
        index = label.index("(")
    except ValueError:
        return label
    return label[:index]


def _write_debug_graph(root: _Node, filename: str):
    graph = nx.MultiDiGraph()
    stack = Stack()
    stack.push(root)
    while stack:
        node = stack.pop()
        graph.add_node(id(node), label=_make_label(node))
        for child in node.children:
            graph.add_edge(id(node), id(child))
            stack.push(child)
    nx.drawing.nx_pydot.write_dot(graph, filename)


write_debug_graph = _write_debug_graph


@dataclasses.dataclass(frozen=True)
class _Context:
    lhs: Optional[SmartBits]
    root: _Node

    def max_width(self, *widths) -> int:
        max_width = max(widths)
        if self.lhs is None:
            return max_width
        return max(max_width, len(self.lhs))


@dataclasses.dataclass
class _Partition:
    context: _Context
    nodes: Set[_Node] = dataclasses.field(default_factory=set, init=False)

    def add(self, node: _Node):
        self.nodes.add(node)


class _PartitionSet:
    def __init__(self):
        self._partitions = {}  # Dict[_Context, _Partition]
        self._context_lookup = {}  # Dict[_Node, _Context]

    def add(self, context: _Context, node: _Node):
        partition = self._partitions.setdefault(context, _Partition(context))
        partition.add(node)
        self._context_lookup[node] = context

    def get_nodes(self, context: _Context) -> Set[_Node]:
        return self._partitions[context].nodes

    def get_partition(self, node: _Node) -> _Partition:
        context = self._context_lookup[node]
        return self._partitions[context]

    def partitions(self) -> Iterable[_Partition]:
        return self._partitions.values()

    def debug_strings(self) -> Iterable[str]:
        for partition in self._partitions.values():
            yield f"Context: {str(partition.context)}"
            for node in partition.nodes:
                yield f"    {str(node)}"


class _PartitionGraph(_Visitor):
    def __init__(self, partitions: _PartitionSet):
        self._partitions = partitions

    def visit(self, node: _Node, context: _Context):
        self._partitions.add(context, node)
        super().visit(node, context)

    def visit_SmartComparisonOp(self, node: _Node, context: _Context):
        super().generic_visit(node, _Context(None, node))

    def visit_SmartShiftOp(self, node: _Node, context: _Context):
        self.visit(node.children[0], context)
        self.visit(node.children[1], _Context(None, node.children[1]))

    def visit_SmartExtendOp(self, node: _Node, context: _Context):
        for child in node.children:
            self.visit(child, _Context(None, child))

    def visit_SmartConcatOp(self, node: _Node, context: _Context):
        for child in node.children:
            self.visit(child, _Context(None, child))

    def visit_SmartMuxOp(self, node: _Node, context: _Context):
        for child in node.children:
            self.visit(child, _Context(None, child))

    def visit_SmartReductionOp(self, node: _Node, context: _Context):
        super().generic_visit(node, _Context(None, node))


def _max_width(partition: _Partition) -> int:
    leaves = filter(lambda n: len(n.children) == 0, partition.nodes)
    widths = (len(leaf.expr.bits) for leaf in leaves)
    return partition.context.max_width(*widths)


class _ResultWidthDeterminator(_Visitor):
    def __init__(self, partitions: _PartitionSet, widths: Dict[_Node, int]):
        self._partitions = partitions
        self._widths = widths

    def visit_SmartBitsExpr(self, node: _Node):
        self._widths[node] = len(node.expr.bits)

    def visit_SmartNAryContextualOp(self, node: _Node):
        super().generic_visit(node)
        partition = self._partitions.get_partition(node)
        self._widths[node] = _max_width(partition)

    def visit_SmartComparisonOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = 1

    def visit_SmartShiftOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = self._widths[node.children[0]]

    def visit_SmartConcatOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = sum(self._widths[child] for child in node.children)

    def visit_SmartMuxOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = max(self._widths[child] for child in node.children)

    def visit_SmartReductionOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = 1

    def visit_SmartSignedOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = self._widths[node.children[0]]

    def visit_SmartExtendOp(self, node: _Node):
        super().generic_visit(node)
        self._widths[node] = self._widths[node.children[0]]


class _ResultSignednessDeterminator(_Visitor):
    def __init__(self, signednesses: Dict[_Node, bool]):
        self._signednesses = signednesses

    def visit_SmartNAryContextualOp(self, node: _Node):
        super().generic_visit(node)
        signedness = all(self._signednesses[child] for child in node.children)
        self._signednesses[node] = signedness

    def visit_SmartComparisonOp(self, node: _Node):
        super().generic_visit(node)
        self._signednesses[node] = False

    def visit_SmartBitsExpr(self, node: _Node):
        self._signednesses[node] = issigned(node.expr.bits)

    def visit_SmartReductionOp(self, node: _Node):
        super().generic_visit(node)
        self._signednesses[node] = False

    def visit_SmartSignedOp(self, node: _Node):
        super().generic_visit(node)
        self._signednesses[node] = node.expr.op.signed

    def visit_SmartExtendOp(self, node: _Node):
        super().generic_visit(node)
        self._signednesses[node] = self._signednesses[node.children[0]]


class _SignednessInserter(_Transformer):
    def __init__(
            self,
            widths: Dict[_Node, int],
            signednesses: Dict[_Node, bool],
    ):
        self._widths = widths
        self._signednesses = signednesses

    def _insert_signed_op(self, node: _Node, signedness: bool) -> _Node:
        expr = SmartSignedOp(signedness, None)
        new_node = _Node(expr, [node])
        self._widths[new_node] = self._widths[node]
        self._signednesses[new_node] = signedness
        return new_node

    def visit_SmartNAryContextualOp(self, node: _Node) -> _Node:
        signedness = self._signednesses[node]
        children = [
            self._insert_signed_op(self.visit(child), signedness)
            for child in node.children
        ]
        node.update(children)
        return node

    def generic_visit(self, node: _Node) -> _Node:
        children = [
            self._insert_signed_op(self.visit(child), self._signednesses[child])
            for child in node.children
        ]
        node.update(children)
        return node


class _PushDownExtensions(_Transformer):
    def __init__(self, widths, signednesses):
        self._widths = widths
        self._signednesses = signednesses

    def _maybe_extend(self, child: _Node, width: int) -> _Node:
        diff = width - self._widths[child]
        if diff <= 0:
            return child
        signedness = self._signednesses[child]
        node = _Node(SmartExtendOp(diff, signedness, None), [child])
        self._widths[node] = width
        return node

    def visit_SmartNAryContextualOp(self, node: _Node) -> _Node:
        width = self._widths[node]
        children = [
            self._maybe_extend(self.visit(child), width)
            for child in node.children
        ]
        node.update(children)
        return node

    def visit_SmartComparisonOp(self, node: _Node) -> _Node:
        width = max(self._widths[child] for child in node.children)
        children = [
            self._maybe_extend(self.visit(child), width)
            for child in node.children
        ]
        node.update(children)
        return node

    def visit_SmartMuxOp(self, node: _Node) -> _Node:
        width = self._widths[node]
        children = [
            self._maybe_extend(self.visit(child), width)
            for child in node.children[:-1]
        ]
        sel_width = clog2(len(node.children) - 1)
        children.append(self._maybe_extend(node.children[-1], sel_width))
        node.update(children)
        return node


class _ShiftOperandWidthMatcher(_PushDownExtensions):
    def visit_SmartShiftOp(self, node: _Node) -> _Node:
        lchild, rchild = node.children
        lwidth, rwidth = map(self._widths.get, node.children)
        if lwidth == rwidth:
            return node
        if lwidth > rwidth:
            rchild = self._maybe_extend(rchild, lwidth)
            node.update([lchild, rchild])
            return node
        # TODO(rsetaluri): Figure out why this actually works...
        return node


class _Evaluator(_Visitor):
    def visit_SmartBitsExpr(self, node: _Node) -> Bits:
        return node.expr.bits.typed_value()

    def visit_SmartOp(self, node: _Node) -> Bits:
        operands = self.generic_visit(node)
        return node.expr.op(*operands)

    def visit_SmartMuxOp(self, node: _Node) -> Bits:
        operands = self.generic_visit(node)
        sel = operands.pop()
        sel_width = clog2(len(node.children) - 1)
        assert len(sel) >= sel_width
        if len(sel) > sel_width:
            sel = sel[:sel_width]
        return node.expr.op(*operands, sel)


def _determine_result_widths(lhs: SmartBits, root: _Node) -> Dict[_Node, int]:
    context = _Context(lhs, root)
    partitions = _PartitionSet()
    _PartitionGraph(partitions).visit(root, context)
    widths = {}
    _ResultWidthDeterminator(partitions, widths).visit(root)
    return widths


def _determine_result_signednesses(root: _Node) -> Dict[_Node, bool]:
    signednesses = {}
    _ResultSignednessDeterminator(signednesses).visit(root)
    return signednesses


def _insert_signednesses(
        root: _Node,
        widths: Dict[_Node, int],
        signednesses: Dict[_Node, bool],
) -> _Node:
    root = _SignednessInserter(widths, signednesses).visit(root)
    return root


def _push_down_extensions(
        root: _Node,
        widths: Dict[_Node, int],
        signednesses: Dict[_Node, int],
) -> _Node:
    root = _PushDownExtensions(widths, signednesses).visit(root)
    root = _ShiftOperandWidthMatcher(widths, signednesses).visit(root)
    return root


def _evaluate(root: _Node) -> Bits:
    return _Evaluator().visit(root)


def _force_width(value: SmartBits, width: int) -> SmartBits:
    diff = len(value) - width
    if diff == 0:
        return value
    value = value.typed_value()
    if diff > 0:
        value = value[:-diff]
    else:
        value = value.ext(-diff)
    return SmartBits.from_bits(value)


def evaluate_assignment(lhs: SmartBits, rhs: SmartExpr) -> SmartBits:
    root = make_tree(rhs)
    widths = _determine_result_widths(lhs, root)
    signednesses = _determine_result_signednesses(root)
    root = _insert_signednesses(root, widths, signednesses)
    root = _push_down_extensions(root, widths, signednesses)
    result = SmartBits.from_bits(_evaluate(root))
    return _force_width(result, len(lhs))



import dataclasses

from magma.backend.mlir.mlir import MlirDialect, begin_dialect, end_dialect
from magma.backend.mlir.mlir import MlirAttribute
from magma.backend.mlir.mlir import MlirOp, MlirRegion, MlirBlock
from magma.backend.mlir.mlir import MlirType
from magma.backend.mlir.mlir_printer_utils import print_attr_dict
from magma.backend.mlir.printer_base import PrinterBase


builtin = MlirDialect("builtin")
begin_dialect(builtin)


@dataclasses.dataclass(frozen=True)
class IntegerType(MlirType):
    n: int

    def emit(self) -> str:
        return f"i{self.n}"


@dataclasses.dataclass(frozen=True)
class IntegerAttr(MlirAttribute):
    value: int

    def emit(self) -> str:
        return str(self.value)


@dataclasses.dataclass
class ModuleOp(MlirOp):
    def __post_init__(self):
        self._block = self.new_region().new_block()

    @property
    def block(self) -> MlirBlock:
        return self._block

    def add_operation(self, operation: MlirOp):
        self._block.add_operation(operation)

    def print_op(self, printer: PrinterBase):
        printer.print("module")
        if self.attr_dict:
            printer.print(" attributes ")
            print_attr_dict(self.attr_dict, printer)


end_dialect()



from typing import Union
from magma.bit import Bit
from magma.logging import root_logger
from magma.primitives.mux import mux
from magma.t import Type, In, Out, Direction
from magma.protocol_type import MagmaProtocol
from magma.tuple import Product, ProductKind


_logger = root_logger()


class ReadyValidException(Exception):
    pass


def _check_T(cls, T):
    if not issubclass(T, (Type, MagmaProtocol)):
        raise TypeError(
            f"{cls}[T] expected T to be a subclass of m.Type or"
            f" m.MagmaProtocol not {T}"
        )
    undirected_T = T.undirected_t
    if undirected_T is not T:
        _logger.warning(
            f"Type {T} used with ReadyValid is not undirected, converting"
            " to undirected type"
        )
    return undirected_T


def _maybe_add_data(cls, fields, T, qualifier):
    if T is None:
        return
    T = _check_T(cls, T)
    fields["data"] = qualifier(T)


class HandShakeError(Exception):
    pass


class EnqError(HandShakeError):
    def __init__(self, cls):
        super().__init__(f"{cls} does not support enq")


class NoEnqError(HandShakeError):
    def __init__(self, cls):
        super().__init__(f"{cls} does not support no_enq")


class DeqError(HandShakeError):
    def __init__(self, cls):
        super().__init__(f"{cls} does not support deq")


class NoDeqError(HandShakeError):
    def __init__(self, cls):
        super().__init__(f"{cls} does not support no_deq")


def _enq_error(self, value, when=True):
    raise EnqError(type(self))


def _no_enq_error(self, value, when=True):
    raise NoEnqError(type(self))


def _deq_error(self, value, when=True):
    raise DeqError(type(self))


def _no_deq_error(self, value, when=True):
    raise NoDeqError(type(self))


class HandShakeKind(ProductKind):
    """Undirected handshake type."""
    _qualifers_ = {
        "output": lambda x: x,
        "input": lambda x: x,
        "data": lambda x: x
    }

    def _make_fields(cls, T):
        """
        Each HandShake implementation should define a dict of _field_names_ for
        the signals (e.g. "ready" and "valid").  For consistency, we define the
        output/input names in terms of a Producer (i.e. ReadyValid outputs a
        valid and inputs a ready).

        Each implementation should define a dict of _qualifiers_ used to
        qualify the ports for producer, consumer, etc...
        """
        fields = {
            cls._field_names_["handshake_output"]:
                cls._qualifers_["output"](Bit),
            cls._field_names_["handshake_input"]:
                cls._qualifers_["input"](Bit)
        }
        _maybe_add_data(cls, fields, T, cls._qualifers_["data"])
        return fields

    def __getitem__(cls, T: Union[Type, MagmaProtocol]):
        assert cls._field_names_ is not None, "Undefined field name impl"
        fields = cls._make_fields(T)

        bases = (cls, )
        if hasattr(cls, "Base_T"):
            # Qualified variants include the unqualified type as a base class.
            # e.g. Producer(ReadyValid[Bits[8]]) will be a subtype of
            # ReadyValid[Bits[8]].
            bases += (cls.Base_T[T], )
        name = f"{cls}[{T}]"

        # Populates namespace for proper cache behavior (based on
        # hwtypes.adt_meta logic).
        ns = {}
        ns['__module__'] = cls.__module__
        name = f'{cls}[{T}]'
        ns['__qualname__'] = name

        # _cache_handler is implemented in hwtypes.adt_meta.
        return cls._cache_handler(True, fields, name, bases, ns)

    def __new__(mcs, name, bases, namespace, fields=None, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace, fields, **kwargs)

        if not cls._field_names_:
            # Skip for base class (variants already exist and pointer is not
            # needed).
            return cls

        variants = (HandShakeMonitor, HandShakeDriver, HandShakeConsumer,
                    HandShakeProducer)
        if issubclass(cls, variants):
            return cls  # these don't need variants

        # Metaprogram variants for base type (e.g. for ReadyValid[Bits[8]],
        # create Producer(ReadyValid[Bits[8]])).
        for variant in variants:
            # Strip "HandShake" prefix for string name.
            suffix = f"{variant.__name__}"[len("HandShake"):]

            # Make variant type with appropriate base classes.
            variant_T = type(f"{suffix}({cls.__name__})",
                             (variant, cls), {})

            # Store pointer to variant (used by qualifiers).
            setattr(cls, f"{suffix}_T", variant_T)
            variant_T.Base_T = cls  # store pointer to base type

        return cls

    def flip(cls):
        raise TypeError(f"Cannot flip {cls.__name__} (undirected)")

    def qualify(cls, direction):
        if direction is Direction.In:
            return Monitor(cls)
        if direction is Direction.Out:
            return Driver(cls)
        if direction is Direction.Undirected:
            return Undirected(cls)
        raise TypeError(f"Cannot qualify {cls.__name__} with {direction}")

    @property
    def undirected_data_t(cls):
        try:
            data = cls.field_dict["data"]
        except KeyError:
            return None
        return data.undirected_t

    def __str__(cls):
        return cls.__name__


class HandShake(Product, metaclass=HandShakeKind):
    _field_names_ = None  # must be defined by implementation

    def fired(self):
        """
        Returns a bit that is high when ready and valid are both high.
        """
        raise NotImplementedError


class HandShakeProducerKind(HandShakeKind):
    _qualifers_ = {"output": Out, "input": In, "data": Out}

    def __getitem__(cls, T: Union[Type, MagmaProtocol]):
        t = super().__getitem__(T)
        if T is None:
            # Remove deq/no_deq methods if no data.
            t.deq = _deq_error
            t.no_deq = _no_deq_error
        return t

    def flip(cls):
        return Consumer(cls)


class HandShakeConsumerKind(HandShakeKind):
    _qualifers_ = {"output": In, "input": Out, "data": In}

    def __getitem__(cls, T: Union[Type, MagmaProtocol]):
        t = super().__getitem__(T)
        if T is None:
            t.enq = _enq_error
            t.no_enq = _no_enq_error
        return t

    def flip(cls):
        return Producer(cls)


class HandShakeConsumer(HandShake, metaclass=HandShakeConsumerKind):
    def fired(self):
        if not self.valid.driven():
            raise ReadyValidException(
                "Cannot invoke fired on consumer when valid isn't driven"
            )

        return self.ready & self.valid.value()

    def enq(self, value, when=True):
        """
        Produces the data `value` when `when` is high.

        Uses "last connect semantics", which allows multiple `enq` methods to
        be invoked with different conditions with precedence given to the later
        invocations.
        """
        if when is not True:
            if not (self.valid.driven() and self.data.driven()):
                raise ReadyValidException(
                    "Cannot use enq when without default driver"
                )
            curr_valid = self.valid.value()
            when |= curr_valid
            self.valid.unwire(curr_valid)

            curr_data = self.data.value()
            value = mux([curr_data, value], when)
            self.data.unwire(curr_data)
        self.valid @= when
        self.data @= value
        return value

    def no_enq(self, when=True):
        """
        Indicates no enqueue occurs.  Valid is set to false, and bits are
        connected to 0. (TODO: Should be an unitialized wire (Don't Care))
        """
        if when is not True:
            if not (self.valid.driven() and self.data.driven()):
                raise ReadyValidException(
                    "Cannot use no_deq with condition without an existing"
                    "valid and data driver"
                )
            valid = self.valid.value()
            data = self.data.value()
            self.valid.unwire(valid)
            self.data.unwire(data)
            self.valid @= mux([valid, False], when)
            self.data @= mux([data, 0], when)
        else:
            self.valid @= False
            self.data @= 0


class HandShakeProducer(HandShake, metaclass=HandShakeProducerKind):
    def fired(self):
        if not self.ready.driven():
            raise ReadyValidException(
                "Cannot invoke fired on producer when ready isn't driven"
            )
        return self.ready.value() & self.valid

    def deq(self, when=True):
        """
        Assert ready on this port and return associated data when `when` is
        high.

        This is typically used when valid has been asserted on the producer
        side.
        """
        if when is not True:
            if not self.ready.driven():
                raise ReadyValidException(
                    "Cannot use deq with when without a default ready driver"
                )
            ready = self.ready.value()
            self.ready.unwire(ready)
            self.ready @= ready | when
        else:
            self.ready @= when
        return self.data

    def no_deq(self, when=True):
        """
        Indicates no dequeue occurs, ready is set to False.
        """
        if when is not True:
            if not self.ready.driven():
                raise ReadyValidException(
                    "Cannot use no_deq with condition without an existing"
                    " ready driver"
                )
            ready = self.ready.value()
            self.ready.unwire(ready)
            self.ready @= mux([ready, False], when)
        else:
            self.ready @= False


class HandShakeMonitorKind(HandShakeKind):
    _qualifers_ = {"output": In, "input": In, "data": In}

    def flip(cls):
        return Driver(cls)


class HandShakeMonitor(HandShake, metaclass=HandShakeMonitorKind):
    pass


class HandShakeDriverKind(HandShakeKind):
    _qualifers_ = {"output": Out, "input": Out, "data": Out}

    def flip(cls):
        return Monitor(cls)


class HandShakeDriver(HandShake, metaclass=HandShakeDriverKind):
    pass


class ReadyValid(HandShake):
    """
    An I/O Bundle containing 'valid' and 'ready' signals that handshake
    the transfer of data stored in the 'data' subfield.

    The base protocol is undirected and is meant to be used with the modifiers:
    * m.Consumer(m.ReadyValid[T])  # data/valid are inputs, ready is output
    * m.Producer(m.ReadyValid[T])  # data/valid are outputs, ready is input

    T is the type of data to be wrapped in Ready/Valid.

    To create a ReadyValid type with no data, use T=None.  In this case, there
    will be no data port.

    The actual semantics of ready/valid are enforced via the use of concrete
    subclasses.
    """
    _field_names_ = {"handshake_output": "valid", "handshake_input": "ready"}


class Decoupled(ReadyValid):
    """
    `valid` indicates the prdoucer has put valid data in `data`, and `ready`
    indicates that the consumer is ready to accept the data this cycle.  No
    requirements are made on the signaling of ready or valid.
    """
    # TODO: Add down convert logic from Irrevocable consumer to Decoupled,
    # dropping guarantees of irrevocability.  Cannot be used on producer
    pass


class EnqIO(Decoupled.Producer_T):
    """
    Alias for Producer(Decoupled[T]).

    drives (outputs) `valid` and `data`, inputs `ready`.
    """
    pass


class DeqIO(Decoupled.Consumer_T):
    """
    Alias for Consumer(Decoupled[T]).

    drives (outputs) `ready`, inputs `valid` and `data`.
    """
    pass


class Irrevocable(ReadyValid):
    """
    Promises not to change the value of `data` after a cycle where `valid` is
    high and `ready` is low.  Additionally, once `valid` is raised, it will
    never be lowered until after `ready` has also been raised.
    """
    # TODO: Add upconvert logic from Decoupled producer to Irrevocable (can use
    # Irrevocable where a Decoupled is expected). Cannot be used on consumer.
    pass


class CreditValid(HandShake):
    _field_names_ = {"handshake_output": "valid", "handshake_input": "credit"}


def Consumer(T: HandShakeKind):
    if not issubclass(T, HandShake):
        raise TypeError(f"Consumer({T}) is unsupported")
    undirected_T = T.undirected_data_t
    return T.Consumer_T[undirected_T]


def Producer(T: HandShakeKind):
    if not issubclass(T, HandShake):
        raise TypeError(f"Producer({T}) is unsupported")
    undirected_T = T.undirected_data_t
    return T.Producer_T[undirected_T]


def Monitor(T: HandShakeKind):
    if not issubclass(T, HandShake):
        raise TypeError(f"Monitor({T}) is unsupported")
    undirected_T = T.undirected_data_t
    return T.Monitor_T[undirected_T]


def Driver(T: HandShakeKind):
    if not issubclass(T, HandShake):
        raise TypeError(f"Driver({T}) is unsupported")
    undirected_T = T.undirected_data_t
    return T.Driver_T[undirected_T]


def Undirected(T: HandShakeKind):
    if not issubclass(T, HandShake):
        raise TypeError(f"Undirected({T}) is unsupported")
    if not hasattr(T, 'Base_T'):
        return T  # already undirected
    undirected_T = T.undirected_data_t
    return T.Base_T[undirected_T]


def is_producer(T):
    return isinstance(T, (HandShakeProducerKind, HandShakeProducer))


def is_consumer(T):
    return isinstance(T, (HandShakeConsumerKind, HandShakeConsumer))


# TODO: QueueIO and Queue



try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache

cachedFunctions = []


def clear_cachedFunctions():
    for func in cachedFunctions:
        func.cache_clear()


def cache_definition(fn):
    cachedFunctions.append(lru_cache(maxsize=None)(fn))
    return cachedFunctions[-1]

from magma.protocol_type import MagmaProtocolMeta, MagmaProtocol

# wires
from .wire import wire, unwire

# types
from .t import *

from .bit import *
import magma.bit_primitives

from .array import *
from .bits import *
from .bfloat import BFloat
from .tuple import *
from .clock import *
from .clock_io import ClockIO
from .interface import *
from .ref import LazyDefnRef, LazyInstRef

# circuit
from .circuit import *

# higher-order operators
from .braid import *

# verilog
from .frontend.verilog import (declare_from_verilog, declare_from_verilog_file,
                               define_from_verilog, define_from_verilog_file,
                               DeclareFromVerilog, DeclareFromVerilogFile,
                               DefineFromVerilog, DefineFromVerilogFile,
                               set_verilog_importer)
from .backend.verilog import *

# compile
from .compile import *

from .logging import root_logger


mantle_target = None
def set_mantle_target(t):
     global mantle_target
     if mantle_target is not None and mantle_target != t:
         root_logger().warning('changing mantle target', mantle_target, t )
     mantle_target = t

from .backend.util import set_codegen_debug_info
from .enum import Enum
import magma.util
import magma.syntax

from .is_primitive import isprimitive
from .is_definition import isdefinition

from .uniquification import UniquificationPass


from hwtypes.bit_vector_abc import TypeFamily

from magma.family import get_family
_Family_ = get_family()

from .generator import Generator, Generator2, DebugGenerator2
from .monitor import MonitorIO, MonitorGenerator, make_monitor_ports
from .inline_verilog import inline_verilog
from .display import display, posedge, negedge, File, time
from .log import info, debug, warning, error


from .syntax.sequential2 import sequential2
from .syntax.combinational2 import combinational2
from .syntax.inline_combinational import inline_combinational
from .syntax.coroutine import coroutine

from magma.conversions import (
    bit,
    clock, reset, enable, asyncreset, asyncresetn,
    array,
    bits, uint, sint,
    tuple_, namedtuple, product,
    concat, repeat,
    sext, zext, sext_to, sext_by, zext_to, zext_by,
    replace,
    as_bits, from_bits
)
from magma.primitives import (
    LUT, Mux, mux, dict_lookup, list_lookup,
    Register, get_slice, set_slice, slice,
    Memory, set_index, register, Wire,
    AbstractRegister,
)

from magma.types import (BitPattern, Valid, ReadyValid, Consumer, Producer,
                         Decoupled, EnqIO, DeqIO, Irrevocable, Monitor,
                         is_producer, is_consumer, CreditValid)
import magma.smart
from magma.compile_guard import compile_guard, compile_guard_select
from magma.inline_verilog_expression import InlineVerilogExpression
from magma.set_name import set_name
from magma.stubify import (
    circuit_stub, stubify, CircuitStub, zero_stubbifier,
    no_override_driven_zero_stubbifier
)
from magma.compile import MagmaCompileException
from magma.linking import link_module, link_default_module, clear_link_info
import magma.math
from magma.when import when, elsewhen, otherwise
from magma.value_utils import fill
from magma.bind2 import bind2, make_bind_ports

################################################################################
# BEGIN ALIASES
################################################################################
Reg = Register
################################################################################
# END ALIASES
################################################################################



from typing import Dict, List, Optional, Union

from magma.circuit import DefineCircuitKind, CircuitKind, CircuitType
from magma.generator import Generator2Kind, Generator2
from magma.passes.passes import DefinitionPass, pass_lambda
from magma.view import PortView
from magma.wire_utils import wire_value_or_port_view
from magma.t import Type, In


_BOUND_INSTANCE_INFO_KEY = "_bound_instance_info_"
_BOUND_GENERATOR_INFO_KEY = "_bound_generator_info_"

DutType = Union[DefineCircuitKind, Generator2Kind]
BindModuleType = Union[CircuitKind, Generator2Kind]
ArgumentType = Union[Type, PortView]


def set_bound_instance_info(inst: CircuitType, info: Dict):
    setattr(inst, _BOUND_INSTANCE_INFO_KEY, info)


def get_bound_instance_info(inst: CircuitType) -> Dict:
    return getattr(inst, _BOUND_INSTANCE_INFO_KEY, None)


def is_bound_instance(inst: CircuitType) -> bool:
    return get_bound_instance_info(inst) is not None


def get_bound_generator_info(inst: CircuitType) -> List[Generator2Kind]:
    try:
        info = getattr(inst, _BOUND_GENERATOR_INFO_KEY)
    except AttributeError:
        info = list()
        setattr(inst, _BOUND_GENERATOR_INFO_KEY, info)
    return info


class BindGenerators(DefinitionPass):
    def __call__(self, defn):
        if not isinstance(defn, Generator2):
            return
        gen = type(defn)
        for bind_generator in get_bound_generator_info(gen):
            bind_module = bind_generator(defn, *defn._args_, **defn._kwargs_)
            bind_args = getattr(bind_module, "bind2_args", list())
            bind2(defn, bind_module, *bind_args)


bind_generators = pass_lambda(BindGenerators)


def make_bind_ports(defn_or_decl: CircuitKind) -> Dict[str, Type]:
    return {
        name: In(type(port))
        for name, port in defn_or_decl.interface.ports.items()
    }


def _bind_generator_impl(dut: Generator2Kind, bind_module: Generator2Kind):
    info = get_bound_generator_info(dut)
    info.append(bind_module)


def _bind_impl(
        dut: DefineCircuitKind,
        bind_module: CircuitKind,
        args: List[ArgumentType],
        compile_guard: str,
):
    arguments = list(dut.interface.ports.values()) + args
    with dut.open():
        inst = bind_module()
        for param, arg in zip(inst.interface.ports.values(), arguments):
            wire_value_or_port_view(param, arg)
        info = {"args": args, "compile_guard": compile_guard}
        set_bound_instance_info(inst, info)


def bind2(
        dut: DutType,
        bind_module: BindModuleType,
        *args,
        compile_guard: Optional[str] = None,
):
    args = list(args)
    are_generators = (
        isinstance(dut, Generator2Kind) and
        isinstance(bind_module, Generator2Kind)
    )
    if are_generators:
        if args:
            raise ValueError(
                "Expected no arguments for binding generators. "
                "Implement bind_arguments() instead."
            )
        return _bind_generator_impl(dut, bind_module)
    are_modules = (
        isinstance(dut, DefineCircuitKind) and
        isinstance(dut, CircuitKind)
    )
    if are_modules:
        return _bind_impl(dut, bind_module, args, compile_guard)
    raise TypeError(dut, bind_module)



_PRIMITIVE_TO_PYTHON_MAP = {
    "and": "and_",
    "or": "or_",
    "xor": "xor",
    "shl": "lshift",
    "lshr": "rshift",
    "ashr": "rshift",
    "urem": "mod",
    "srem": "mod",
    "udiv": "floordiv",
    "sdiv": "floordiv",
    "ule": "le",
    "ult": "lt",
    "uge": "ge",
    "ugt": "gt",
    "sle": "le",
    "slt": "lt",
    "sge": "ge",
    "sgt": "gt",
    "not": "invert"
}


def primitive_to_python(key):
    return _PRIMITIVE_TO_PYTHON_MAP.get(key, key)



import io
import sys
from typing import Optional

from magma.backend.mlir.compile_to_mlir_opts import CompileToMlirOpts
from magma.backend.mlir.printer_base import PrinterBase
from magma.backend.mlir.translation_unit import TranslationUnit
from magma.circuit import DefineCircuitKind


def compile_to_mlir(
        top: DefineCircuitKind,
        sout: Optional[io.TextIOBase] = None,
        opts: CompileToMlirOpts = CompileToMlirOpts()):
    if sout is None:
        sout = sys.stdout
    translation_unit = TranslationUnit(top, opts)
    translation_unit.compile()
    printer = PrinterBase(sout=sout)
    translation_unit.mlir_module.print(printer)



import contextlib
import hashlib
import string
from typing import Mapping, Optional, Union

from ast_tools.stack import _SKIP_FRAME_DEBUG_STMT, get_symbol_table

from magma.array import Array
from magma.bit import Bit
from magma.circuit import Circuit
from magma.clock import ClockTypes
from magma.definition_context import DefinitionContext, get_definition_context
from magma.digital import Digital
from magma.interface import IO
from magma.passes.passes import CircuitPass
from magma.primitives.wire import Wire
from magma.ref import DefnRef, InstRef, ArrayRef, TupleRef
from magma.t import Type, Direction, In
from magma.tuple import Tuple
from magma.view import PortView, InstView
from magma.wire_utils import wire_value_or_driver, WiringError


ValueLike = Union[Type, PortView]


class InlineVerilogError(RuntimeError):
    pass


def _get_view_inst_parent(view):
    while not isinstance(view, InstView):
        assert isinstance(view, PortView), type(view)
        return _get_view_inst_parent(view.parent)
    return view


def _make_inline_value(
        inline_value_map: Mapping[str, ValueLike], value: ValueLike) -> str:
    if isinstance(value, Array) and not issubclass(value.T, Digital):
        key = ", ".join(
            _make_inline_value(inline_value_map, t)
            for t in reversed(value)
        )
        return f"'{{{key}}}"
    if isinstance(value, Tuple):
        raise NotImplementedError(value)
    key = f"__magma_inline_value_{len(inline_value_map)}"
    inline_value_map[key] = value
    return f"{{{key}}}"


def _build_io(inline_value_map: Mapping[str, ValueLike]) -> IO:
    if not inline_value_map:
        # Add dummy port so IO is not empty.
        io = IO(I=In(Bit))
        io.I.unused()
        return io
    io = IO()
    for key, value in inline_value_map.items():
        if isinstance(value, PortView):
            T = value.T
        else:
            T = type(value)
        io += IO(**{key: In(T)})
    return io


def _inline_verilog(
        context: DefinitionContext,
        format_str: str,
        format_args: Mapping[str, ValueLike],
        inline_value_map: Mapping[str, ValueLike],
        inline_wire_prefix: str):
    inline_verilog_modules = context.set_default_metadata(
        "inline_verilog_modules", [])
    format_args = format_args.copy()
    for key, arg in format_args.items():
        if isinstance(arg, (Type, PortView)):
            arg = _make_inline_value(inline_value_map, arg)
        format_args[key] = arg
    inline_str = format_str.format(**format_args)

    # Because modules/instances are sorted lexigraphically in the generated
    # verilog, in order for the inline verilog in the output to be in
    # statement-order, we need to generate a lexicographically increasing string
    # based on the order of the inline verilog statement in the python. A trick
    # to do this with numbers is, for every 10 modules, to insert a prefix 9 so
    # that it comes after the previous 1-9 digits.
    i = len(inline_verilog_modules)
    prefix = "9" * (i // 10)
    suffix = i % 10

    class _InlineVerilog(Circuit):
        name = f"{context.placer.name}_inline_verilog_{prefix}{suffix}"
        io = _build_io(inline_value_map)
        # Since each inline value (key) maps to a port, we populate the
        # connect_references dictionary with a map from key to port.
        connect_references = {}
        for key in inline_value_map:
            port = getattr(io, key)
            port.unused()  # these are needed so CoreIR knows it's a definition
            connect_references[key] = port
        inline_verilog_strs = [(inline_str, connect_references)]

    inline_verilog_modules.append(_InlineVerilog)
    inst_name = f"{context.placer.name}_inline_verilog_inst_{prefix}{suffix}"
    inst = _InlineVerilog(name=inst_name)
    # If there no interpolated values, _build_io adds a dummy port to avoid an
    # empty IO. We drive that with 0 here to avoid a hanging input.
    if not inline_value_map:
        inst.I @= 0

    for key, value in inline_value_map.items():
        try:
            wire_value_or_driver(getattr(inst, key), value)
        except WiringError:
            raise InlineVerilogError(
                f"Found reference to undriven input port: {repr(value)}"
            ) from None


def _process_fstring_syntax(
        format_str: str,
        format_args: Mapping[str, ValueLike],
        inline_value_map: Mapping[str, ValueLike],
        symbol_table: Mapping) -> str:
    fieldnames = [
        fname
        for _, fname, _, _ in string.Formatter().parse(format_str)
        if fname
    ]
    for field in fieldnames:
        if field in format_args:
            continue
        try:
            value = eval(field, {}, symbol_table)
        except NameError:
            continue
        if isinstance(value, (Type, PortView)):
            # Magma values (more precisely value-like objects, i.e. values or
            # PortView objects) are interpolated into strings specially. We
            # handle that here by creating a wire instance to make sure the
            # values persist.
            value = _make_inline_value(inline_value_map, value)
            # Stage for subsequent format call for regular kwargs inside
            # _inline_verilog.
            value = value.replace("{", "{{").replace("}", "}}")
        format_str = format_str.replace(f"{{{field}}}", str(value))
    return format_str


def _process_inline_verilog(
        context: DefinitionContext,
        format_str: str,
        format_args: Mapping[str, ValueLike],
        symbol_table: Mapping,
        inline_wire_prefix: str):
    inline_value_map = {}
    if symbol_table is not None:
        format_str = _process_fstring_syntax(
            format_str, format_args, inline_value_map, symbol_table)
    _inline_verilog(
        context, format_str, format_args, inline_value_map, inline_wire_prefix)


def inline_verilog_impl(
        format_str: str,
        format_args: Mapping[str, ValueLike],
        symbol_table: Mapping,
        inline_wire_prefix: str):
    context = get_definition_context()
    _process_inline_verilog(
        context, format_str, format_args, symbol_table, inline_wire_prefix)


def inline_verilog(
        format_str: str,
        inline_wire_prefix: str = "_magma_inline_wire",
        **kwargs):
    exec(_SKIP_FRAME_DEBUG_STMT)
    format_args = kwargs
    symbol_table = get_symbol_table([inline_verilog], copy_locals=True)
    inline_verilog_impl(
        format_str, format_args, symbol_table, inline_wire_prefix)


class ProcessInlineVerilogPass(CircuitPass):
    def __call__(self, _):
        pass



import ctypes
import io
import struct

import pytest

import env
from pybind11_tests import ConstructorStats
from pybind11_tests import buffers as m

np = pytest.importorskip("numpy")


def test_from_python():
    with pytest.raises(RuntimeError) as excinfo:
        m.Matrix(np.array([1, 2, 3]))  # trying to assign a 1D array
    assert str(excinfo.value) == "Incompatible buffer format!"

    m3 = np.array([[1, 2, 3], [4, 5, 6]]).astype(np.float32)
    m4 = m.Matrix(m3)

    for i in range(m4.rows()):
        for j in range(m4.cols()):
            assert m3[i, j] == m4[i, j]

    cstats = ConstructorStats.get(m.Matrix)
    assert cstats.alive() == 1
    del m3, m4
    assert cstats.alive() == 0
    assert cstats.values() == ["2x3 matrix"]
    assert cstats.copy_constructions == 0
    # assert cstats.move_constructions >= 0  # Don't invoke any
    assert cstats.copy_assignments == 0
    assert cstats.move_assignments == 0


# https://foss.heptapod.net/pypy/pypy/-/issues/2444
# TODO: fix on recent PyPy
@pytest.mark.xfail(
    env.PYPY, reason="PyPy 7.3.7 doesn't clear this anymore", strict=False
)
def test_to_python():
    mat = m.Matrix(5, 4)
    assert memoryview(mat).shape == (5, 4)

    assert mat[2, 3] == 0
    mat[2, 3] = 4.0
    mat[3, 2] = 7.0
    assert mat[2, 3] == 4
    assert mat[3, 2] == 7
    assert struct.unpack_from("f", mat, (3 * 4 + 2) * 4) == (7,)
    assert struct.unpack_from("f", mat, (2 * 4 + 3) * 4) == (4,)

    mat2 = np.array(mat, copy=False)
    assert mat2.shape == (5, 4)
    assert abs(mat2).sum() == 11
    assert mat2[2, 3] == 4 and mat2[3, 2] == 7
    mat2[2, 3] = 5
    assert mat2[2, 3] == 5

    cstats = ConstructorStats.get(m.Matrix)
    assert cstats.alive() == 1
    del mat
    pytest.gc_collect()
    assert cstats.alive() == 1
    del mat2  # holds a mat reference
    pytest.gc_collect()
    assert cstats.alive() == 0
    assert cstats.values() == ["5x4 matrix"]
    assert cstats.copy_constructions == 0
    # assert cstats.move_constructions >= 0  # Don't invoke any
    assert cstats.copy_assignments == 0
    assert cstats.move_assignments == 0


def test_inherited_protocol():
    """SquareMatrix is derived from Matrix and inherits the buffer protocol"""

    matrix = m.SquareMatrix(5)
    assert memoryview(matrix).shape == (5, 5)
    assert np.asarray(matrix).shape == (5, 5)


def test_pointer_to_member_fn():
    for cls in [m.Buffer, m.ConstBuffer, m.DerivedBuffer]:
        buf = cls()
        buf.value = 0x12345678
        value = struct.unpack("i", bytearray(buf))[0]
        assert value == 0x12345678


def test_readonly_buffer():
    buf = m.BufferReadOnly(0x64)
    view = memoryview(buf)
    assert view[0] == 0x64
    assert view.readonly
    with pytest.raises(TypeError):
        view[0] = 0


def test_selective_readonly_buffer():
    buf = m.BufferReadOnlySelect()

    memoryview(buf)[0] = 0x64
    assert buf.value == 0x64

    io.BytesIO(b"A").readinto(buf)
    assert buf.value == ord(b"A")

    buf.readonly = True
    with pytest.raises(TypeError):
        memoryview(buf)[0] = 0
    with pytest.raises(TypeError):
        io.BytesIO(b"1").readinto(buf)


def test_ctypes_array_1d():
    char1d = (ctypes.c_char * 10)()
    int1d = (ctypes.c_int * 15)()
    long1d = (ctypes.c_long * 7)()

    for carray in (char1d, int1d, long1d):
        info = m.get_buffer_info(carray)
        assert info.itemsize == ctypes.sizeof(carray._type_)
        assert info.size == len(carray)
        assert info.ndim == 1
        assert info.shape == [info.size]
        assert info.strides == [info.itemsize]
        assert not info.readonly


def test_ctypes_array_2d():
    char2d = ((ctypes.c_char * 10) * 4)()
    int2d = ((ctypes.c_int * 15) * 3)()
    long2d = ((ctypes.c_long * 7) * 2)()

    for carray in (char2d, int2d, long2d):
        info = m.get_buffer_info(carray)
        assert info.itemsize == ctypes.sizeof(carray[0]._type_)
        assert info.size == len(carray) * len(carray[0])
        assert info.ndim == 2
        assert info.shape == [len(carray), len(carray[0])]
        assert info.strides == [info.itemsize * len(carray[0]), info.itemsize]
        assert not info.readonly


def test_ctypes_from_buffer():
    test_pystr = b"0123456789"
    for pyarray in (test_pystr, bytearray(test_pystr)):
        pyinfo = m.get_buffer_info(pyarray)

        if pyinfo.readonly:
            cbytes = (ctypes.c_char * len(pyarray)).from_buffer_copy(pyarray)
            cinfo = m.get_buffer_info(cbytes)
        else:
            cbytes = (ctypes.c_char * len(pyarray)).from_buffer(pyarray)
            cinfo = m.get_buffer_info(cbytes)

        assert cinfo.size == pyinfo.size
        assert cinfo.ndim == pyinfo.ndim
        assert cinfo.shape == pyinfo.shape
        assert cinfo.strides == pyinfo.strides
        assert not cinfo.readonly



from mako.template import Template

from magma.circuit import DeclareCircuit, DefineCircuit, EndDefine
from magma.common import deprecated
from magma.config import config, RuntimeConfig
from magma.frontend.dummy_verilog_importer import DummyVerilogImporter
from magma.frontend.verilog_importer import ImportMode


_pyverilog_importer_initialized = False
# We register a runtime configuration for allowing dynamically setting the
# verilog importer. By default, it is set to a dummy importer so that pyverilog
# is only imported on demand.
config._register(verilog_importer=RuntimeConfig(DummyVerilogImporter()))


def _init_pyverilog_importer():
    global _pyverilog_importer_initialized
    if _pyverilog_importer_initialized:
        return
    if not isinstance(config.verilog_importer, DummyVerilogImporter):
        return
    from magma.frontend.pyverilog_importer import PyverilogImporter
    config.verilog_importer = PyverilogImporter({})
    _pyverilog_importer_initialized = True


def _from_verilog(source, func, *, target_modules=None, type_map={}):
    _init_pyverilog_importer()
    importer = config.verilog_importer
    importer.reset(type_map=type_map)
    mode = ImportMode.DECLARE if func is DeclareCircuit else ImportMode.DEFINE
    importer.import_(source, mode)
    modules = importer._magma_modules.copy()
    if target_modules is None:
        return modules

    def _filter(mod):
        return mod.name in target_modules

    return list(filter(_filter, modules))


def _from_verilog_file(file, func, *, target_modules=None, type_map={}):
    if file is None:
        return None
    with open(file) as f:
        verilog = f.read()
    circuits = _from_verilog(verilog, func, target_modules=target_modules,
                             type_map=type_map)
    # Store the original verilog file name, currently used by m.compile to
    # generate a .sv when compiling a circuit that was defined from a verilog
    # file.
    for circuit in circuits:
        circuit.verilog_file_name = file
    return circuits


def set_verilog_importer(importer):
    """Convenience function for setting the runtime verilog importer"""
    config.verilog_importer = importer


def declare_from_verilog(source, *, target_modules=None, type_map={}):
    return _from_verilog(source, DeclareCircuit, target_modules=target_modules,
                         type_map=type_map)


def declare_from_verilog_file(file, *, target_modules=None, type_map={}):
    return _from_verilog_file(file, DeclareCircuit,
                              target_modules=target_modules,
                              type_map=type_map)


def define_from_verilog(source, *, target_modules=None, type_map={}):
    return _from_verilog(source, DefineCircuit,
                         target_modules=target_modules,
                         type_map=type_map)


def define_from_verilog_file(file, *, target_modules=None, type_map={}):
    return _from_verilog_file(file, DefineCircuit,
                              target_modules=target_modules,
                              type_map=type_map)


# The following functions call their analagous functions above, formatted as
# snake_case rathern than CamelCase. These functions are marked as deprecated.
@deprecated(msg="Use _from_verilog instead")
def FromVerilog(*args, **kwargs):
    return _from_verilog(*args, **kwargs)


@deprecated(msg="Use _from_verilog_file instead")
def FromVerilogFile(*args, **kwargs):
    return _from_verilog_file(*args, **kwargs)


@deprecated(msg="Use declare_from_verilog instead")
def DeclareFromVerilog(*args, **kwargs):
    return declare_from_verilog(*args, **kwargs)


@deprecated(msg="Use declare_from_verilog_file instead")
def DeclareFromVerilogFile(*args, **kwargs):
    return declare_from_verilog_file(*args, **kwargs)


@deprecated(msg="Use define_from_verilog instead")
def DefineFromVerilog(*args, **kwargs):
    return define_from_verilog(*args, **kwargs)


@deprecated(msg="Use define_from_verilog_file instead")
def DefineFromVerilogFile(*args, **kwargs):
    return define_from_verilog_file(*args, **kwargs)


# The functions (Declare|Define)FromTemplatedVerilog(|File) are now deprecated.
@deprecated
def FromTemplatedVerilog(templatedverilog, func, type_map, **kwargs):
    verilog = Template(templatedverilog).render(**kwargs)
    return FromVerilog(verilog, func, type_map)


@deprecated
def FromTemplatedVerilogFile(file, func, type_map, **kwargs):
    if file is None:
        return None
    templatedverilog = open(file).read()
    return FromTemplatedVerilog(templatedverilog, func, type_map, **kwargs)


@deprecated
def DeclareFromTemplatedVerilog(source, type_map={}, **kwargs):
    return FromTemplatedVerilog(source, DeclareCircuit, type_map, **kwargs)


@deprecated
def DeclareFromTemplatedVerilogFile(file, type_map={}, **kwargs):
    return FromTemplatedVerilogFile(file, DeclareCircuit, type_map, **kwargs)


@deprecated
def DefineFromTemplatedVerilog(source, type_map={}, **kwargs):
    return FromTemplatedVerilog(source, DefineCircuit, type_map, **kwargs)


@deprecated
def DefineFromTemplatedVerilogFile(file, type_map={}, **kwargs):
    return FromTemplatedVerilogFile(file, DefineCircuit, type_map, **kwargs)



import abc
import contextlib
import dataclasses
import functools
import itertools
from typing import Any, Iterable, Optional, Set, Tuple, Union

from magma.ref import get_ref_inst


@functools.lru_cache(None)
def _abstract_register_type():
    # NOTE(rsetaluri): Circular dependency.
    from magma.primitives.register import AbstractRegister
    return AbstractRegister


@functools.lru_cache(None)
def _enable_type():
    # NOTE(rsetaluri): Circular dependency.
    from magma.clock import Enable
    return Enable


class WhenSyntaxError(SyntaxError):
    pass


class NoPrecedingWhenError(WhenSyntaxError):
    def __init__(self, type):
        super().__init__(f"Cannot use {type} without a previous when")


class ElsewhenWithoutPrecedingWhenError(NoPrecedingWhenError):
    def __init__(self):
        super().__init__("elsewhen")


class OtherwiseWithoutPrecedingWhenError(NoPrecedingWhenError):
    def __init__(self):
        super().__init__("otherwise")


@dataclasses.dataclass(frozen=True)
class ConditionalWire:
    drivee: Any
    driver: Any


class _BlockBase(contextlib.AbstractContextManager):
    def __init__(self, parent: Optional['_WhenBlock']):
        self._parent = parent
        self._children = list()
        self._conditional_wires = list()
        self._default_drivers = dict()

    def spawn(self, info: '_WhenBlockInfo') -> '_WhenBlock':
        child = _WhenBlock(self, info)
        self._children.append(child)
        return child

    def get_conditional_wires_for_drivee(self, i):
        return list(
            filter(lambda x: x.drivee is i, self._conditional_wires)
        )

    def remove_conditional_wire(self, i):
        self._conditional_wires = list(
            filter(lambda x: x.drivee is not i, self._conditional_wires)
        )
        if self.root._default_drivers.pop(i, None) is not None:
            self.root.builder.remove_default_driver(i)

    def add_conditional_wire(self, i, o):
        self._conditional_wires.append(ConditionalWire(i, o))
        builder = self.root.builder
        if i.driven() and i not in builder.output_to_index:
            driver = i.value()
            # Keep wired when contexts because in this case, the driver could
            # have been wired in a different when context (so the default value
            # comes from the result of a previous when context)
            i.unwire(driver, keep_wired_when_contexts=True)
            self.root.add_default_driver(i, driver)
        self.root.builder.add_drivee(i)
        self.root.builder.add_driver(o)

    @property
    @abc.abstractmethod
    def root(self) -> '_WhenBlock':
        raise NotImplementedError()

    @abc.abstractmethod
    def new_elsewhen_block(self, info: '_ElseWhenBlockInfo'):
        raise NotImplementedError()

    @abc.abstractmethod
    def new_otherwise_block(self):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def condition(self) -> Optional:
        raise NotImplementedError()

    @abc.abstractmethod
    def elsewhen_blocks(self) -> Iterable['_BlockBase']:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def otherwise_block(self) -> Optional['_BlockBase']:
        raise NotImplementedError()

    def conditional_wires(self) -> Iterable[ConditionalWire]:
        yield from self._conditional_wires

    def default_drivers(self) -> Iterable[ConditionalWire]:
        if self.root is not self:
            raise RuntimeError("Can only get default drivers from root")
        return (
            ConditionalWire(k, v)
            for k, v in self._default_drivers.items()
        )

    def get_default_drivers_dict(self) -> Iterable[ConditionalWire]:
        if self.root is not self:
            raise RuntimeError("Can only get default drivers from root")
        return self._default_drivers

    def get_default_driver(self, drivee) -> Any:
        if self.root is not self:
            raise RuntimeError("Can only get default driver from root")
        return self._default_drivers[drivee]

    def add_default_driver(self, drivee, driver):
        if self.root is not self:
            raise RuntimeError("Can only add default driver to root")
        self._default_drivers[drivee] = driver
        self.root.builder.add_default_driver(drivee, driver)

    def children(self) -> Iterable['_BlockBase']:
        yield from self._children

    def __enter__(self):
        self.root.builder.add_condition(self.condition)
        _set_curr_block(self)
        return self

    def _add_reg_enables(self):
        """
        If we wire the input of a register, check if there is an enable port
        and wire to 1 if found.

        This will overwrite an existing wiring if the user already has done it.

        The warning raised will indicate to the user that the explicit wiring
        is unnecessary and promote the idiom of not explicitly wiring enables.
        """
        conditional_wires = list(self._conditional_wires)
        # We need to keep track of Enable values we've wired and make sure to
        # only wire them once. Otherwise, an elaborated register input value may
        # be wired multiple times.
        # NOTE(leonardt): If we wanted to allow users to wire Enable values
        # explicitly, we could lookup whether a wire to enable exists in
        # conditional_wires (which would also avoid multiple wirings).
        seen = set()
        for wire in conditional_wires:
            inst = get_ref_inst(wire.drivee.name)
            if (
                    inst is None
                    or not isinstance(type(inst), _abstract_register_type())
            ):
                continue
            ce_port = type(inst).get_enable(inst)
            if (
                    ce_port is None
                    # Explicitly wired in this block, skip.
                    or self.get_conditional_wires_for_drivee(ce_port)
                    or (
                        ce_port.driven()
                        and not ce_port.driven_implicitly_by_when
                    )
                    or ce_port in seen
            ):
                continue
            ce_port.wire(1, driven_implicitly_by_when=True)
            seen.add(ce_port)

    @abc.abstractmethod
    def _get_exit_block(self):
        """Returns the block that should be set as the current block when
        __exit__ is called.
        """
        raise NotImplementedError()

    def __exit__(self, exc_type, exc_value, traceback):
        self._add_reg_enables()
        _set_curr_block(self._get_exit_block())
        _set_prev_block(self)


BlockBase = _BlockBase


def get_all_blocks(
        block: _BlockBase,
        include_self: bool = True,
) -> Iterable[_BlockBase]:
    if include_self:
        yield block
    sub_blocks = itertools.chain(
        block.children(),
        block.elsewhen_blocks(),
        () if block.otherwise_block is None else (block.otherwise_block,),
    )
    for sub_block in sub_blocks:
        yield from get_all_blocks(sub_block)


@dataclasses.dataclass
class _WhenBlockInfo:
    condition: Any  # TODO(rsetaluri): Should actually be bit


@dataclasses.dataclass
class _ElseWhenBlockInfo(_WhenBlockInfo):
    pass


class _WhenBlock(_BlockBase):
    def __init__(self, parent: Optional[_BlockBase], info: _WhenBlockInfo):
        super().__init__(parent)
        self._info = info
        self._elsewhens = list()
        self._otherwise = None
        self._builder = None
        if self._parent is None:
            from magma.primitives.when import WhenBuilder
            self._builder = WhenBuilder(self)

    def new_elsewhen_block(self, info: '_ElseWhenBlockInfo'):
        block = _ElseWhenBlock(self, info)
        self._elsewhens.append(block)
        return block

    def new_otherwise_block(self):
        if self._otherwise is not None:
            raise WhenSyntaxError()
        block = _OtherwiseBlock(self)
        self._otherwise = block
        return block

    @property
    def builder(self) -> Any:
        if self._builder is None:
            raise AttributeError()
        return self._builder

    @property
    def root(self) -> '_WhenBlock':
        if self._parent is None:
            return self
        return self._parent.root

    @property
    def condition(self) -> Any:
        return self._info.condition

    def elsewhen_blocks(self) -> Iterable['_BlockBase']:
        yield from self._elsewhens

    @property
    def otherwise_block(self) -> Optional['_BlockBase']:
        return self._otherwise

    def __enter__(self):
        this = _BlockBase.__enter__(self)
        assert this is self
        _set_prev_block(None)
        return self

    def _get_exit_block(self):
        return self._parent


class _ElseWhenBlock(_BlockBase):
    def __init__(self, parent: Optional[_BlockBase], info: _ElseWhenBlockInfo):
        super().__init__(parent)
        self._info = info

    def new_elsewhen_block(self, info: '_ElseWhenBlockInfo'):
        return self._parent.new_elsewhen_block(info)

    def new_otherwise_block(self):
        return self._parent.new_otherwise_block()

    @property
    def root(self) -> _WhenBlock:
        return self._parent.root

    @property
    def condition(self) -> Any:
        return self._info.condition

    def elsewhen_blocks(self) -> Iterable['_BlockBase']:
        yield from ()

    @property
    def otherwise_block(self) -> None:
        return None

    def _get_exit_block(self):
        return self._parent._parent


class _OtherwiseBlock(_BlockBase):
    def new_elsewhen_block(self, _):
        raise ElsewhenWithoutPrecedingWhenError()

    def new_otherwise_block(self):
        raise OtherwiseWithoutPrecedingWhenError()

    @property
    def root(self) -> _WhenBlock:
        return self._parent.root

    @property
    def condition(self) -> None:
        return None

    def elsewhen_blocks(self) -> Iterable['_BlockBase']:
        yield from ()

    @property
    def otherwise_block(self) -> None:
        return None

    def _get_exit_block(self):
        return self._parent._parent


_curr_block = None
_prev_block = None


@contextlib.contextmanager
def no_when():
    block = _get_curr_block()
    _set_curr_block(None)
    yield
    _set_curr_block(block)


@contextlib.contextmanager
def temp_when(temp_ctx):
    block = _get_curr_block()
    _set_curr_block(temp_ctx)
    yield
    _set_curr_block(block)


def _get_curr_block() -> Optional[_BlockBase]:
    global _curr_block
    return _curr_block


def _set_curr_block(curr_block: Optional[_BlockBase]):
    global _curr_block
    _curr_block = curr_block


def _reset_curr_block():
    _set_curr_block(None)


def _get_prev_block() -> Optional[_BlockBase]:
    global _prev_block
    return _prev_block


def _set_prev_block(prev_block: Optional[_BlockBase]):
    global _prev_block
    _prev_block = prev_block


def _reset_prev_block():
    _set_prev_block(None)


get_curr_block = _get_curr_block


def _reset_context():
    _reset_curr_block()
    _reset_prev_block()


reset_context = _reset_context


def _get_else_block(block: _BlockBase) -> Optional[_BlockBase]:
    if isinstance(block, _OtherwiseBlock):
        return None
    if isinstance(block, _ElseWhenBlock):
        # TODO(rsetaluri): We could augment _ElseWhenBlock to keep track of its
        # index in the parent block, or alternatively keep an explicit pointer
        # to the next else/otherwise block in each _ElseWhenBlock.
        parent_elsewhen_blocks = list(block._parent.elsewhen_blocks())
        index = parent_elsewhen_blocks.index(block) + 1
        if index == len(parent_elsewhen_blocks):
            return block._parent.otherwise_block
        return parent_elsewhen_blocks[index]
    if isinstance(block, _WhenBlock):
        try:
            else_block = next(block.elsewhen_blocks())
        except StopIteration:
            pass
        else:
            return else_block
        return block.otherwise_block


_Op = Union[ConditionalWire, _BlockBase]


def _get_then_ops(block: _BlockBase) -> Iterable[_Op]:
    yield from block.conditional_wires()
    yield from block.children()


def _get_else_ops(block: _BlockBase) -> Iterable[_Op]:
    if isinstance(block, _OtherwiseBlock):
        return _get_then_ops(block)
    if isinstance(block, _ElseWhenBlock):
        return (block,)
    raise TypeError(block)


def _get_assignees_and_latches(ops: Iterable[_Op]) -> Tuple[Set, Set]:
    """Compute the assigned values and latches implied by a block.

    Note: this is a conservative check that requires values are assigned in
    every case.  For example, there may be an unreachable case where a value is
    not assigned, but that requires symbolic evaluation to discover.  In
    general, symbolic evaluation is difficult, but may be viable in the
    hardware domain where we may have less issues with path explosion,
    aliasing, arrays, etc...

    Parameters
    ----------
    ops : Iterable[_Op]
        A sequence of ops (assignment or if-then-else block)

    Returns
    -------
    Set[Type]
        The values assigned in this block by any operation.

    Set[Type]
        The latches implied by this block, i.e. values which are not assigned in
        all branches.
    """
    assignees, latches = set(), set()
    for op in ops:
        if isinstance(op, ConditionalWire):
            assignees.add(op.drivee)
            continue
        if isinstance(op, _BlockBase):
            then_ops = _get_then_ops(op)
            then_assignees, then_latches = _get_assignees_and_latches(
                then_ops
            )
            else_block = _get_else_block(op)
            if else_block is None:
                latches.update(then_latches.difference(assignees))
                latches.update(then_assignees.difference(assignees))
                continue
            else_ops = _get_else_ops(else_block)
            else_assignees, else_latches = _get_assignees_and_latches(
                else_ops
            )
            latches.update(then_latches.difference(assignees))
            latches.update(else_latches.difference(assignees))
            latches.update(
                (
                    then_assignees
                    .symmetric_difference(else_assignees)
                    .difference(assignees)
                )
            )
            assignees.update(then_assignees.union(else_assignees))
            # Filter out any latches that were created in this when context.
            latches = set(filter(
                lambda x: x.enclosing_when_context is not op,
                latches
            ))
            continue
        raise TypeError(op)
    return assignees, latches


def find_inferred_latches(block: _BlockBase) -> Set:
    if not (isinstance(block, _WhenBlock) and block.root is block):
        raise TypeError("Can only find inferred latches on root when block")
    ops = tuple(block.default_drivers()) + (block,)
    _, latches = _get_assignees_and_latches(ops)
    return latches


def when(cond):
    info = _WhenBlockInfo(cond)
    curr_block = _get_curr_block()
    if curr_block is None:
        return _WhenBlock(None, info)
    return curr_block.spawn(info)


def elsewhen(cond):
    info = _ElseWhenBlockInfo(cond)
    prev_block = _get_prev_block()
    if prev_block is None:
        raise ElsewhenWithoutPrecedingWhenError()
    return prev_block.new_elsewhen_block(info)


def otherwise():
    prev_block = _get_prev_block()
    if prev_block is None:
        raise OtherwiseWithoutPrecedingWhenError()
    return prev_block.new_otherwise_block()



from inspect import getouterframes, currentframe
from pathlib import PurePath

from magma.backend import verilog, blif, firrtl, dot
from magma.backend.coreir.coreir_compiler import CoreIRCompiler
from magma.backend.mlir.mlir_compiler import MlirCompiler
from magma.bind import BindPass
from magma.bind2 import bind_generators
from magma.compiler import Compiler
from magma.compile_exception import MagmaCompileException
from magma.config import get_compile_dir
from magma.inline_verilog import ProcessInlineVerilogPass
from magma.is_definition import isdefinition
from magma.passes.clock import WireClockPass
from magma.passes.drive_undriven import drive_undriven
from magma.passes.instance_callback_pass import instance_callback_pass
from magma.passes.terminate_unused import terminate_unused
from magma.uniquification import (
    UniquificationMode,
    uniquification_pass,
    reset_names,
)


__all__ = ["compile"]


def _make_compiler(output, main, basename, opts):
    if output == "verilog":
        return verilog.VerilogCompiler(main, basename)
    if output == "blif":
        return blif.BlifCompiler(main, basename)
    if output == "firrtl":
        return firrtl.FirrtlCompiler(main, basename)
    if output == "dot":
        return dot.DotCompiler(main, basename)
    if output == "coreir":
        return CoreIRCompiler(main, basename, opts)
    if output == "coreir-verilog":
        opts["output_verilog"] = True
        return CoreIRCompiler(main, basename, opts)
    if output == "mlir":
        return MlirCompiler(main, basename, opts)
    if output == "mlir-verilog":
        opts["output_verilog"] = True
        return MlirCompiler(main, basename, opts)
    raise NotImplementedError(f"Backend '{output}' not supported")


def _get_basename(basename):
    if get_compile_dir() == "callee_file_dir":
        callee_filename = getouterframes(currentframe())[2][1]
        callee_dir = PurePath(callee_filename).parent
        return str(callee_dir.joinpath(basename))
    return basename


def _partial_kw(fn, **specialized_kwargs):

    def func(*args, **kwargs):
        return fn(*args, **kwargs, **specialized_kwargs)

    return func


def _make_bind_compile_fn(output: str):
    output_is_mlir = (output == "mlir" or output == "mlir-verilog")
    if not output_is_mlir:
        return compile
    return _partial_kw(compile, use_packed_arrays=True)


def compile(basename, main, output="coreir-verilog", **kwargs):
    if not isdefinition(main):
        raise MagmaCompileException(
            f"Trying to compile empty definition {main}")
    if hasattr(main, "circuit_definition"):
        main = main.circuit_definition
    basename = _get_basename(basename)
    opts = kwargs.copy()
    compiler = _make_compiler(output, main, basename, opts)

    bind_generators(main)

    # Default behavior is to perform uniquification, but can be overriden.
    original_names = uniquification_pass(main, opts.get("uniquify", "UNIQUIFY"))

    # Steps to process inline verilog generation. Required to be run after
    # uniquification.
    ProcessInlineVerilogPass(main).run()

    # Bind after uniquification so the bind logic works on unique modules.
    # NOTE(rsetaluri): This is a hack to use packed arrays when the compilation
    # goes through MLIR.
    compile_fn = _make_bind_compile_fn(output)
    BindPass(main, compile_fn, opts.get("user_namespace"),
             opts.get("verilog_prefix")).run()

    if opts.get("drive_undriven", False):
        drive_undriven(main)
    if opts.get("terminate_unused", False):
        terminate_unused(main)

    instance_callback_pass_data = instance_callback_pass(main)

    result = compiler.compile()
    result = {} if result is None else result

    if hasattr(main, "fpga"):
        main.fpga.constraints(basename)

    result["instance_callback_pass_data"] = instance_callback_pass_data

    reset_names(original_names)

    return result



from typing import Union

from magma.t import Type
from magma.view import PortView
from magma.primitives.xmr import XMRSink, XMRSource


ValueOrPortView = Union[Type, PortView]


class WiringError(Exception):
    pass


def wire_value_or_driver(drivee: Type, driver: Type):
    assert drivee.is_input()
    if driver.is_input():
        value = driver.value()
        if value is None:
            raise WiringError(drivee, driver)
        drivee @= value
    elif driver.is_mixed():
        for drivee_i, driver_i in zip(drivee, driver):
            wire_value_or_driver(drivee_i, driver_i)
    else:
        drivee @= driver


def wire_value_or_port_view(drivee: Type, value_or_port_view: ValueOrPortView):
    if isinstance(value_or_port_view, PortView):
        defn = value_or_port_view.parent.inst.defn
        with defn.open():
            xmr_sink = XMRSink(value_or_port_view)()
            wire_value_or_driver(xmr_sink.I, value_or_port_view.port)
        xmr_source = XMRSource(value_or_port_view)()
        wire_value_or_driver(drivee, xmr_source.O)
        return
    wire_value_or_driver(drivee, value_or_port_view)



import dataclasses
import io
import os
import pathlib
import subprocess
import sys
from typing import List, Optional

from magma.backend.mlir.common import try_call
from magma.backend.mlir.errors import MlirCompilerError
from magma.config import config, EnvConfig
from magma.logging import root_logger


config._register(circt_home=EnvConfig("CIRCT_HOME", None))

_logger = root_logger().getChild("mlir_backend")


class MlirToVerilogError(MlirCompilerError):
    pass


@dataclasses.dataclass
class MlirToVerilogOpts:
    pass


def _circt_home() -> Optional[pathlib.Path]:
    """Returns the CIRCT_HOME config if it is not None. Otherwise returns None.
    """
    if config.circt_home is None:
        return None
    return pathlib.Path(config.circt_home).resolve()


def _circt_opt_binary(circt_home: Optional[pathlib.Path]) -> str:
    """Returns the absolute path to the circt-opt binary if @circt_home is not
    None. Otherwise, returns just the binary name (i.e. looks for the binary in
    PATH).
    """
    if circt_home is not None:
        return str(circt_home / "build/bin/circt-opt")
    return "circt-opt"


def _circt_opt_cmd(
        circt_home: pathlib.Path,
        opts: MlirToVerilogOpts,
) -> List[str]:
    bin_ = f"{_circt_opt_binary(circt_home)}"
    passes = [
        "--lower-seq-to-sv",
        "--canonicalize",
        "--hw-cleanup",
        "--prettify-verilog",
        "--export-verilog",
    ]
    extra_opts = [
        "-o=/dev/null",
    ]
    return [bin_] + passes + extra_opts


def _run_subprocess(args, stdin, stdout) -> int:
    stdin_pipe = (
        try_call(lambda: stdin.fileno(), io.UnsupportedOperation) is None
    )
    stdout_pipe = (
        try_call(lambda: stdout.fileno(), io.UnsupportedOperation) is None
    )
    stdin_actual = subprocess.PIPE if stdin_pipe else stdin
    stdout_actual = subprocess.PIPE if stdout_pipe else stdout
    proc = subprocess.Popen(args, stdin=stdin_actual, stdout=stdout_actual)
    if stdin_pipe:
        proc.stdin.write(stdin.read())
        proc.stdin.close()
    proc.wait()
    if stdout_pipe:
        stdout.write(proc.stdout.read())
    return proc.returncode


def _make_stream(filename, mode, default):
    if filename is None:
        return default, False
    return open(filename, mode), True


def circt_opt_binary_exists() -> bool:
    circt_home = _circt_home()
    circt_opt_binary = _circt_opt_binary(circt_home)
    # NOTE(rsetaluri): We could simply check for the existence of the file in
    # the filesystem, but instead we actually run the binary and (a) check that
    # it exists (by catching FileNotFoundError), and (b) that it is executable
    # with a basic '--help' interface. Also, temporary buffers are passed for
    # stdin/stdout since we do not care to access them.
    try:
        returncode = _run_subprocess(
            [circt_opt_binary, "--help"],
            stdin=io.BytesIO(),
            stdout=io.BytesIO(),
        )
    except FileNotFoundError:
        return False
    return returncode == 0


def mlir_to_verilog(
        istream: io.RawIOBase,
        ostream: io.RawIOBase = sys.stdout,
        opts: MlirToVerilogOpts = MlirToVerilogOpts(),
):
    circt_home = _circt_home()
    cmd = _circt_opt_cmd(circt_home, opts)
    _logger.debug(f"Running cmd: {' '.join(cmd)}")
    returncode = _run_subprocess(cmd, istream, ostream)
    if returncode != 0:
        cmd_str = " ".join(cmd)
        raise MlirToVerilogError(
            f"Error running {cmd_str}, got returncode {returncode}"
        )



from __future__ import print_function
from .python_simulator import PythonSimulator
from ..passes.debug_name import DebugNamePass
from ..is_definition import isdefinition
from ..circuit import CircuitType, EndCircuit
from ..scope import *
from ..array import Array
from ..bit import Bit
from ..bitutils import seq2int, int2seq
from ..ref import InstRef, DefnRef
from ..compatibility import builtins
from magma.waveform import waveform
from code import compile_command
import re
import sys
if sys.platform == "linux" or sys.platform == "linux2" or sys.platform == "darwin":
    import readline
import cmd

__all__ = ['simulate', 'SimulationConsole']

class DisplayExpr:
    idx = 0
    def __init__(self, bit, string, scope):
        self.object = bit
        self.scope = scope
        self.string = string

        DisplayExpr.idx += 1
        self.idx = DisplayExpr.idx

    def display(self, value):
        print("\t{}: {} = {}".format(self.idx, self.string, value))

def print_err(str):
    print(str, file=sys.stderr)

def split_index(str):
    r = re.compile(r"^([a-zA-Z]+)((?:\[[0-9]+\])+)$")
    match = r.match(str)
    if match is None:
        return None

    idxr = re.compile(r"\[([0-9]+)\]")
    idx_match = idxr.findall(match[2])

    return match[1], [int(i) for i in idx_match]

def convert_to_bools(val, bit):
    if isinstance(val, int):
        return int2seq(val, len(bit))
    elif isinstance(val, bool):
        return val
    elif isinstance(val, list):
        newval = []
        for i,v in enumerate(val):
            newval.append(convert_to_bools(v, bit[i]))
        return newval

def describe_instance(inst):
    desc_str = type(inst).__name__ + ": "
    if inst.decl is not None:
        desc_str += inst.decl.varname + " @ " + inst.decl.filename + ":" + str(inst.decl.lineno)

    desc_str += " (" + inst.name + ")"
    return desc_str

def describe_interface(interface):
    print("\nInterface Inputs:")
    for name, bit in interface.ports.items():
        if bit.is_output():
            if isinstance(bit, Array):
                print("  Bit[" + str(len(bit)) + "]:" + name)
            else:
                print("  Bit: " + name)

    print("\nInterface Outputs:")
    for name, bit in interface.ports.items():
        if bit.is_input():
            if isinstance(bit, Array):
                print("  Bit[" + str(len(bit)) + "]:" + name)
            else:
                print("  Bit: " + name)

    print("")

def get_bit_full_name(bit):
    name = bit.name
    if isinstance(name, InstRef):
        return name.inst.name + "." + name.name
    elif isinstance(name, DefnRef):
        return str(name.defn) + '.' + name.name
    elif isinstance(name, ArrayRef):
        arrayname = get_bit_full_name(name.array)
        return arrayname + "[" + str(name.index) + "]"
    else:
        return ""

def format_val(val, bit, raw):
    # Is an array of arrays?
    if isinstance(val, list) and len(val) > 0 and isinstance(val[0], list):
        s = '['
        for i, v in enumerate(val):
            s += str(format_val(v, bit[i], raw))
            if i < len(val) - 1:
                s += ", "
        s += ']'
        return s
    else:
        if raw:
            return "".join(['1' if e else '0' for e in reversed(val)])
        else:
            if not isinstance(bit, Array) or isinstance(val, bool):
                val = [val]
            return seq2int(val)

class SimulationConsoleException(Exception):
    pass

class SimulationConsole(cmd.Cmd):
    def __init__(self, circuit, simulator):
        cmd.Cmd.__init__(self, completekey=None)

        self.scope = Scope()
        self.top_circuit = circuit
        self.simulator = simulator
        self.vars = {}
        self.update_vars()

        self.display_exprs = []

        self.cycles = 0
        self.clock_high = False

        self.aliases = { 'x'  : self.do_examine,
                         'd'  : self.do_descend,
                         'rn' : self.do_reverse_cycle,
                         'rs' : self.do_reverse_step }

        self.update_prompt()

    def print_watchpoints(self, points):
        print("Watchpoint hit:")
        for watchpoint in points:
            print("  " + get_bit_full_name(watchpoint.bit) + ": ", end='')
            self.log_val(watchpoint.bit, watchpoint.scope)

    def update_prompt(self):
        self.prompt = str(self.cycles) + ': ' + self.scope.value() + ' >>> '

    def default(self, line):
        if line == 'EOF':
            print()
            return True

        cmd, arg, line = self.parseline(line)
        if cmd in self.aliases:
            func = [self.aliases[cmd]]
        else:
            func = [getattr(self, n) for n in self.get_names() if n.startswith('do_' + cmd)]

        if len(func) == 1:
            return func[0](arg)
        elif len(func) > 1:
            print_err("Ambiguous command")
        else:
            self.console_evaluate(line)

    def precmd(self, line):
        self.advance_clock = False
        self.reeval = False
        self.skip_half = False
        self.skip_next = 0
        self.stepping = True
        self.reversing = False

        return line

    def reverse_simulator(self):
        if self.clock_high and self.cycles == 0: return

        reversecount = self.skip_next
        if self.skip_half:
            reversecount *= 2
            if not self.clock_high:
                reversecount -= 1

        for i in range(reversecount):
            state = self.simulator.rewind(1)
            self.clock_high = state.clock
            if self.clock_high:
                self.cycles -= 1

            if state.triggered_points:
                self.print_watchpoints(state.triggered_points)
                return False

    def step_simulator(self):
        n = self.skip_next 
        if self.skip_half:
            n *= 2
            if self.clock_high:
                n -= 1

        state = self.simulator.advance(n)
        self.clock_high = state.clock
        self.cycles += state.cycles
        if state.triggered_points:
            self.print_watchpoints(state.triggered_points)

    def continue_simulator(self):
        state = self.simulator.cont()
        self.print_watchpoints(state.triggered_points)
        self.clock_high = state.clock
        self.cycles += state.cycles

    def postcmd(self, stop, line):
        if self.reversing:
            self.reverse_simulator()

        elif self.advance_clock:
            if self.stepping:
                self.step_simulator()
            else:
                self.continue_simulator()

        if self.reeval:
            self.simulator.evaluate(True)

        self.update_prompt()

        for e in self.display_exprs:
            e.display(self.get_formatted_val(e.object, e.scope))

        return stop
    
    def get_formatted_val(self, bit, scope, raw=False):
        val = self.simulator.get_value(bit, scope)
        if val is None:
            return "Doesn't exist"

        return format_val(val, bit, raw)

    def log_val(self, bit, scope, raw=False):
        print(self.get_formatted_val(bit, scope, raw))

    def update_vars(self):
        self.vars.clear()
        self.vars['top'] = self.top_circuit

        if self.scope.inst is None:
            self.vars['self'] = self.top_circuit
        else:
            self.vars['self'] = type(self.scope.inst)

        instances = self.vars['self'].instances
        for inst in instances:
            self.vars[inst.name] = inst
            if inst.decl is not None:
                self.vars[inst.decl.varname] = inst

    def console_evaluate(self, line):
        buf = line
        while True:
            try:
                code = compile_command(buf)
            except Exception as e: 
                print_err("Invalid python: {}".format(e))
                return

            if code is None:
                buf += '\n' + builtins.input("."*(len(self.prompt) - 1) + " ")
            else:
                break
        
        try:
            exec(code, None, self.vars)
        except Exception as e:
            print_err("Failed to execute: {}".format(e))


    def parse_circuit(self, name):
        components = name.split('.')
        if len(components) < 1:
            raise SimulationConsoleException("Need at least instance name")
        
        topname = components.pop(0)
        cur = eval(topname, None, self.vars)

        scope = Scope() if cur == self.top_circuit else self.scope

        for idx, comp_name in enumerate(components):
            defn = type(cur) if isinstance(cur, CircuitType) else cur
            index_match = split_index(comp_name)
            bit_idx = None
            if index_match is not None:
                comp_name = index_match[0]
                bit_idx = index_match[1]

            # Last iteration, check for bit first
            if idx == len(components) - 1 and comp_name in cur.interface.ports:
                if bit_idx is None:
                    return cur.interface.ports[comp_name], scope
                else:
                    port = cur.interface.ports[comp_name]
                    for i in bit_idx:
                        port = port[i]
                    return port, scope

            found = False
            for inst in defn.instances:
                if inst.name == comp_name or (inst.decl is not None and inst.decl.varname == comp_name):
                    # Descend into previous instance's scope
                    if cur != self.top_circuit:
                        scope = Scope(parent=scope, instance=cur)

                    cur = inst
                    found = True
                    break

            if not found:
                raise SimulationConsoleException("Invalid name component '{}'".format(comp_name))

        return cur, scope

    def safe_parse_inst(self, name):
            inst, scope = self.parse_circuit(name)
            if not isinstance(inst, CircuitType):
                raise SimulationConsoleException("not an instance")

            return inst

    def parse_next(self, num):
        self.advance_clock = True
        self.stepping = True

        if num:
            try:
                self.skip_next = int(num)
            except:
                print_err("Please provide an integer")
                self.skip_next = 0

        else:
            self.skip_next = 1

    def parse_print(self, arg, raw):
        if not arg:
            print_err('Please provide an argument')
            return

        try:
            printme, scope = self.parse_circuit(arg)
        except Exception as e:
            print_err("Failed to print: {}".format(e))
            return

        if isinstance(printme, Bit) or isinstance(printme, Array):
            self.log_val(printme, scope, raw)
        elif isinstance(printme, CircuitType):
            inst_desc = describe_instance(printme) + ": "
            print(inst_desc)
            for name, bit in printme.interface.ports.items():
                print("  " + name + ": ", end='')
                self.log_val(bit, scope, raw)
        else:
            print_err("Can only print Bits and circuit instances")

    def do_quit(self, arg):
        'quit: Exit the simulator'
        return True

    def do_next(self, arg):
        'next [N]: Advance the clock for N cycles. N defaults to 1 if not provided.'
        self.skip_half = True
        self.parse_next(arg)

    def do_step(self, arg):
        'step [N]: Toggle the clock N times. N defaults to 1 if not provided.'
        self.skip_half = False
        self.parse_next(arg)

    def do_reverse_step(self, arg):
        'reverse_step N: rewinds the circuit N half cycles'
        self.reversing = True
        self.skip_half = False
        self.parse_next(arg)

    def do_reverse_cycle(self, arg):
        'reverse_cycle N: rewinds the circuit N cycles'
        self.reversing = True
        self.skip_half = True
        self.parse_next(arg)

    def do_examine(self, arg):
        'examine BIT: prints the current value of BIT as an array of booleans. Shortcut: x BIT.'
        self.parse_print(arg, True)

    def do_print(self, arg):
        'print BIT: prints the current value of BIT interpreted as an unsigned integer.'
        self.parse_print(arg, False)

    def do_watch(self, arg):
        'watch BIT [VALUE]: sets a watchpoint on BIT.\nThe simulator will interrupt and return to the console when BIT changes value.\nIf the optional argument VALUE is passed in, the simulator will interrupt only when BIT is equal to VALUE.'
        if not arg:
            print_err('Provide a bit to watch')
            return

        args = arg.split()
        bitname = args[0]

        try:
            watchme, scope = self.parse_circuit(bitname)
        except Exception as e:
            print_err("Failed to watch: {}".format(e))
            return

        if len(args) == 2:
            try:
                value = eval(args[1], None, self.vars)
                if isinstance(value, int):
                    value = [bool(i) for i in int2seq(value, len(watchme))]

                if not isinstance(value, list) or not isinstance(value[0], bool):
                    raise SimulationConsoleException("Invalid watch value")
            except Exception as e:
                print_err("Cannot watch for value {}: {}".format(args[1], e))
                return
        else:
            value = None

        if not isinstance(watchme, Bit) and not isinstance(watchme, Array):
            print_err("Can only watch bits or arrays")

        watch_num = self.simulator.add_watchpoint(watchme, scope, value)
        print('Watchpoint {} on {}'.format(watch_num, arg))

    def do_delete(self, arg):
        'delete N: deletes watchpoint N.'
        if not arg:
            print_err('Please provide a watchpoint number')
            return

        try:
            watch_num = int(arg)
        except:
            print_err("delete requires an integer")
            return

        found = self.simulator.delete_watchpoint(watch_num)
        if not found:
            print_err('No watchpoint number {}'.format(watch_num))

    def do_display(self, arg):
        'display BIT: repeatedly prints the value of BIT each time the simulator stops.'

        if not arg:
            print_err('Provide an expression to display')
            return

        try:
            bit, scope = self.parse_circuit(arg)
            if not isinstance(bit, Bit) and not isinstance(bit, Array):
                raise SimulationConsoleException("Can only display bits or arrays")

            display_expr = DisplayExpr(bit, arg, scope)
        except Exception as e:
            print_err("Invalid argument to display: {}".format(e))
            return

        self.display_exprs.append(display_expr)

    def do_undisplay(self, arg):
        'undisplay N: stops displaying the bit at index N.'
        if not arg:
            print_err('Provide an index to stop displaying')
            return

        try:
            idx = int(arg)
        except:
            print_err("undisplay requires an integer")
            return

        for i,e in enumerate(self.display_exprs):
            if e.idx == idx:
                del self.display_exprs[i]
                return 
        print_err('No display number {}'.format(idx))

    def do_up(self, arg):
        "up: Change to the parent circuit's scope."
        if self.scope.parent is not None:
            self.scope = self.scope.parent
            self.update_vars()
        else:
            print_err("Cannot go up")

    def do_descend(self, arg):
        "descend INSTANCE: update the current scope to be inside INSTANCE."
        try:
            inst, scope = self.parse_circuit(arg)

            if not isdefinition(type(inst)):
                print_err("Cannot descend into primitives")
                return
        except Exception as e:
            print_err("Cannot switch scope to '{}': {}".format(arg, e))
            return

        if isinstance(inst, CircuitType):
            self.scope = Scope(parent=scope, instance=inst)
            self.update_vars()
        elif inst == self.top_circuit:
            self.scope = Scope()
            self.update_vars()
        else:
            print_err("You must provide an instance to descend into")

    def do_info(self, arg):
        """info instances|interface|watchpoints:
     instances [INSTANCE]: Display all the instances in the current scope or in INSTANCE if provided
     interface [INSTANCE]: Display the interface bits of the current scope's circuit or in INSTANE if provided
     watchpoints: Display currently active watchpoints"""
        args = arg.split()
        action = args[0]
        instname = args[1] if len(args) == 2 else None
        if action == 'instances':
            defn = self.vars['self']
            if instname:
                try:
                    inst = self.safe_parse_inst(instname)
                except Exception as e:
                    print_err("Cannot get info on '{}': {}".format(instname, e))
                    return
                defn = type(inst)

                if not isdefinition(defn):
                    print_err("Cannot get instances in '{}' because it is a primitive".format(instname))
                    return

            print("")
            for inst in defn.instances:
                desc_str = "  " + describe_instance(inst)
                print(desc_str)
            print("")

        elif action == 'interface':
            defn = self.vars['self']
            if instname:
                try:
                    inst = self.safe_parse_inst(instname)
                except Exception as e:
                    print_err("Cannot get info on '{}': {}".format(instname, e))
                    return
                defn = type(inst)

            describe_interface(defn.interface)
        elif arg == 'watchpoints':
            print("TODO")
        else:
            print_err("I don't know how to give you info on that")

    def do_continue(self, arg): 
        "continue: continue cycling the simulator's clock and evaluating until a watchpoint is hit."
        self.advance_clock = True
        self.stepping = False

    def do_location(self, arg):
        "location: Print the trace of parent scopes of the current scope."
        scopes = []
        curscope = self.scope
        while curscope is not None:
            scopes.insert(0, curscope)
            curscope = curscope.parent

        print("Scope stack: ")
        for s in scopes:
            print("  " + s.value())

    def do_assign(self, arg):
        "assign BIT NEWVAL: sets BIT to NEWVAL. BIT must be an input to the top level circuit."
        if arg is None:
            print_err('Provide a top level input to change')
            return

        args = arg.split(' ', 1)

        try:
            bit = eval(args[0], None, self.vars)
        except Exception as e:
            print_err("Invalid bit for assignment: {}".format(e))
            return

        if bit not in self.top_circuit.interface.outputs():
            print_err("Can only assign values to inputs in the top level circuit")
            return

        try:
            newval = eval(args[1], None, self.vars)
        except Exception as e:
            print_err("Invalid new value".format(e))
            return

        newval = convert_to_bools(newval, bit)

        self.simulator.set_value(bit, newval, self.scope)
        self.reeval = True

    def do_waveform(self, arg):
        if not arg:
            print_err('Please a provide wire')
            return

        try:
            waveme, scope = self.parse_circuit(arg)
        except Exception as e:
            print_err("Invalid argument for waveform: {}".format(e))
            return

        if not isinstance(waveme, Bit) and not isinstance(waveme, Array):
            print_err("Can only provide waveforms for wires")
            return

        labels = [arg]
        signals = []

        for i in range(self.cycles - 1):
            val = self.simulator.get_value(waveme, scope)
            signals.insert(0, [seq2int(val)])
            self.simulator.rewind(2)

        for i in range(self.cycles - 1):
            self.simulator.step()
            self.simulator.step()

        waveform(signals, labels)

    def run(self):
        self.simulator.evaluate()

        print('Magma Interactive Simulator.   Type help or ? to list commands.')
        while True:
            try:
                self.cmdloop()
                break;
            except KeyboardInterrupt:
                print_err('\nKeyboardInterrupt')

    # For test infra
    def runcmd(self, line):
        line = self.precmd(line)
        stop = self.onecmd(line)
        self.postcmd(stop, line)

def simulate(main, simulator_type=PythonSimulator):
    EndCircuit()
    simulator = simulator_type(main, main.CLK)

    DebugNamePass(main).run()

    console = SimulationConsole(main, simulator)
    console.run()



from magma.array import Array
from magma.clock import AsyncReset, AsyncResetN, Clock, _ClockType
from magma.conversions import as_bits, from_bits, bit, convertbit
from magma.digital import Digital
from magma.generator import Generator2
from magma.passes.passes import DefinitionPass, pass_lambda
from magma.primitives.wire import Wire
from magma.ref import NamedRef, ArrayRef, PortViewRef, InstRef
from magma.t import In, Out
from magma.tuple import Tuple


def _sanitize_name(name):
    return name.replace("[", "_").replace("]", "")\
               .replace("(", "").replace(")", "")\
               .replace(",", "").replace(" ", "_")


class InsertCoreIRWires(DefinitionPass):
    def __init__(self, main):
        super().__init__(main)
        self.seen = set()
        self.wire_map = {}

    def _make_wire(self, driver, value, definition):
        is_bits = (isinstance(driver.name, ArrayRef) and
                   isinstance(driver, Digital) and not driver.name.anon())

        # If this is a member of a bits, emit only one wire for the parent
        # Bits, since it shouldn't be flattened
        if is_bits:
            index = driver.name.index
            driver = driver.name.array

        T = type(driver)
        if driver not in self.wire_map:
            driver_name = driver.name.qualifiedname("_")
            driver_name = _sanitize_name(driver_name)

            # Rename driver so it doesn't conflict with new wire instance name
            if isinstance(driver.name, NamedRef):
                driver.name.name = "_" + driver.name.name

            name = f"{driver_name}"
            wire_inst = Wire(T)(name=name)
            self.wire_map[driver] = wire_inst
        wire_inst = self.wire_map[driver]
        wire_input = wire_inst.I
        wire_output = wire_inst.O

        if is_bits:
            # Select the index off the parent bits wire
            wire_output = wire_output[index]
            # NOTE(leonardt): We don't select off the input and instead wire up
            # the entire bits (this avoids an "undriven" error when only one
            # index of the bits is used)

        if not is_bits and not isinstance(driver, Digital):
            driver = as_bits(driver)
            wire_output = from_bits(T, wire_output)

        # Could be already wired for fanout cases
        if not wire_input.driven():
            wire_input @= driver
        if (isinstance(value, _ClockType) and
                not isinstance(wire_output, type(value))):
            # This mean it was cast by the user (e.g. m.clock(value)), so we
            # need to "recast" the wire output
            wire_output = convertbit(wire_output, type(value))
        value @= wire_output

    def _insert_wire(self, value, definition):
        if value.is_mixed():  # mixed children
            for child, _ in value.connection_iter():
                if not child.is_output():
                    self._insert_wire(child, definition)
            return
        if value in self.seen:
            return  # in the case of inouts, we may see more than once
        self.seen.add(value)
        if not value.driven():
            return  # undriven value, skip wire insertion
        driver = value.value()

        while driver is not None and driver.is_driven_anon_temporary():
            value, driver = driver, driver.value()

        descend = (isinstance(driver, (Array, Tuple)) and
                   driver.anon())
        if descend:
            for child, _ in value.connection_iter():
                self._insert_wire(child, definition)
            return

        if (driver is None or driver.is_output() or driver.is_inout() or
                driver.name.anon()):
            return

        if isinstance(driver.name.root(), InstRef):
            return

        if isinstance(driver.name, PortViewRef):
            # Don't descend into hierarchical connections since we assume this
            # logic will be run in that context in another stage of the pass
            return

        value.unwire(driver)
        with definition.open():
            self._make_wire(driver, value, definition)
            self._insert_wire(driver, definition)

    def __call__(self, definition):
        if getattr(definition, "_coreir_wires_inserted_", False):
            return
        definition._coreir_wires_inserted_ = True
        # Copy instances because inserting wire will append to instances.
        instances_copy = definition.instances.copy()
        for instance in instances_copy:
            for value in instance.interface.ports.values():
                if not value.is_output():
                    self._insert_wire(value, definition)

        # See https://github.com/phanrahan/magma/pull/1157 for more context, we
        # need to run this on specifically the when builder instance, since
        # this hasn't been finalized yet and may trace back to temporary
        # values.
        for builder in definition._context_._builders:
            if builder._finalized:
                continue
            for value in builder._io.inst_ports.values():
                if not value.is_output():
                    self._insert_wire(value, definition)
        for value in definition.interface.ports.values():
            if not value.is_output():
                self._insert_wire(value, definition)


insert_coreir_wires = pass_lambda(InsertCoreIRWires)



from distutils.version import LooseVersion
import multiprocessing
import os
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
import subprocess
import sys


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: " +
                ", ".join(e.name for e in self.extensions))

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(
            os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable]

        cfg = 'Release'
        build_args = ['--config', cfg]

        cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]

        cpu_count = max(2, multiprocessing.cpu_count() // 2)
        build_args += ['--', '-j{0}'.format(cpu_count)]

        python_path = sys.executable
        cmake_args += ['-DPYTHON_EXECUTABLE:FILEPATH=' + python_path]

        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args,
                              cwd=self.build_temp)

        subprocess.check_call(
            ['cmake', '--build', '.'] + build_args,
            cwd=self.build_temp)


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='magma-lang',
    version='2.2.42b1',
    url='https://github.com/phanrahan/magma',
    license='MIT',
    maintainer='Lenny Truong',
    maintainer_email='lenny@cs.stanford.edu',
    description='An embedded DSL for constructing hardware circuits',
    scripts=['bin/magma'],
    packages=[
        "magma",
        "magma.frontend",
        "magma.backend",
        "magma.backend.coreir",
        "magma.backend.mlir",
        "magma.passes",
        "magma.primitives",
        "magma.smart",
        "magma.simulator",
        "magma.syntax",
        "magma.syntax.transforms",
        "magma.ssa",
        "magma.testing",
        "magma.types"
    ],
    ext_modules=[CMakeExtension("magma_debug", 'debug-ext')],
    cmdclass=dict(build_ext=CMakeBuild),
    install_requires=[
        "colorlog",
        "astor",
        "six",
        "dataclasses",
        "mako",
        "pyverilog",
        "numpy",
        "graphviz",
        "coreir>=2.0.151",
        "hwtypes>=1.4.4",
        "ast_tools>=0.0.16",
        "staticfg",
        "networkx",
    ],
    python_requires='>=3.6',
    long_description=long_description,
    long_description_content_type="text/markdown"
)



import networkx as nx
from typing import Any, Callable, Iterable, List, Mapping, Tuple, Union


Graph = nx.MultiDiGraph


def write_to_dot(g: Graph, filename: str):
    nx.drawing.nx_pydot.write_dot(g, filename)



from typing import Any, Dict
import weakref

from magma.backend.mlir.builtin import builtin
from magma.backend.mlir.compile_to_mlir_opts import CompileToMlirOpts
from magma.backend.mlir.hardware_module import HardwareModule
from magma.backend.mlir.mlir import MlirSymbol, push_block
from magma.backend.mlir.scoped_name_generator import ScopedNameGenerator
from magma.circuit import CircuitKind, DefineCircuitKind
from magma.passes import dependencies
from magma.t import Type


def _set_module_attrs(mlir_module: builtin.ModuleOp, opts: CompileToMlirOpts):
    lowering_options = [
        f"locationInfoStyle={opts.location_info_style}",
    ]
    if opts.explicit_bitcast:
        lowering_options.append("explicitBitcast")
    if opts.disallow_expression_inlining_in_ports:
        lowering_options.append("disallowExpressionInliningInPorts")
    if lowering_options:
        mlir_module.attr_dict["circt.loweringOptions"] = (
            f"\"{','.join(lowering_options)}\""
        )


class TranslationUnit:
    def __init__(self, magma_top: DefineCircuitKind, opts: CompileToMlirOpts):
        self._magma_top = magma_top
        self._opts = opts
        self._mlir_module = builtin.ModuleOp()
        _set_module_attrs(self._mlir_module, opts)
        self._hardware_modules = {}
        self._symbol_map = {}
        self._symbol_name_generator = ScopedNameGenerator(
            disallow_duplicates=self._opts.disallow_duplicate_symbols,
        )
        self._bind_files = []
        self._xmr_paths = {}

    @property
    def magma_top(self) -> DefineCircuitKind:
        return self._magma_top

    @property
    def mlir_module(self) -> builtin.ModuleOp:
        return self._mlir_module

    @property
    def xmr_paths(self) -> Dict:
        return self._xmr_paths

    def new_hardware_module(
            self, magma_defn_or_decl: CircuitKind) -> HardwareModule:
        return HardwareModule(magma_defn_or_decl, weakref.ref(self), self._opts)

    def get_hardware_module(
            self, magma_defn_or_decl: CircuitKind) -> HardwareModule:
        key = self._make_key(magma_defn_or_decl)
        return self._hardware_modules[key]

    def set_hardware_module(
            self, magma_defn_or_decl: CircuitKind,
            hardware_module: HardwareModule):
        key = self._make_key(magma_defn_or_decl)
        if key in self._hardware_modules:
            raise ValueError(f"Hardware module '{key}' already exists")
        self._hardware_modules[key] = hardware_module

    def has_hardware_module(
            self, magma_defn_or_decl: CircuitKind) -> bool:
        key = self._make_key(magma_defn_or_decl)
        return key in self._hardware_modules

    def get_mapped_symbol(self, obj: Type) -> MlirSymbol:
        return self._symbol_map[obj]

    def get_or_make_mapped_symbol(self, obj: Any, **kwargs) -> MlirSymbol:
        try:
            return self._symbol_map[obj]
        except KeyError:
            pass
        self._symbol_map[obj] = symbol = self.new_symbol(**kwargs)
        return symbol

    def set_mapped_symbol(self, obj: Any, symbol: MlirSymbol):
        if obj in self._symbol_map:
            raise ValueError(f"{obj} already mapped")
        self._symbol_map[obj] = symbol

    def new_symbol(self, **kwargs) -> MlirSymbol:
        name = self._symbol_name_generator(**kwargs)
        return MlirSymbol(name)

    def add_bind_file(self, filename: str):
        self._bind_files.append(filename)

    def compile(self):
        deps = dependencies(self._magma_top, include_self=True)
        with push_block(self._mlir_module):
            for dep in deps:
                if self.has_hardware_module(dep):
                    continue
                hardware_module = self.new_hardware_module(dep)
                hardware_module.compile()
                if hardware_module.hw_module:
                    self.set_hardware_module(dep, hardware_module)
        self._write_listings_file()

    @staticmethod
    def _make_key(magma_defn_or_decl: CircuitKind) -> str:
        return magma_defn_or_decl.name

    def _write_listings_file(self):
        if not self._bind_files:
            return
        filename = self._opts.basename + "_bind_files.list"
        with open(filename, "w") as f:
            listing = "\n".join(self._bind_files)
            f.write(listing)



import inspect
import collections
from magma.config import get_debug_mode
import magma_debug


debug_info = collections.namedtuple("debug_info",
                                    ["filename", "lineno", "module"])


def get_debug_info(frames_to_skip):
    filename, lineno = magma_debug.get_fn_ln(frames_to_skip)
    return debug_info(filename, lineno, None)


def debug_wire(fn):
    """
    Automatically populates the `debug_info` argument for a wire call if it's
    not already passed as an argument
    """
    def wire(i, o=None, debug_info=None):
        if debug_info is None:
            debug_info = get_debug_info(3)
        return fn(i, o, debug_info)
    return wire


def debug_unwire(fn):
    def unwire(i, o=None, debug_info=None, keep_wired_when_contexts=False):
        if debug_info is None:
            debug_info = get_debug_info(3)
        return fn(i, o, debug_info, keep_wired_when_contexts)
    return unwire
